{****************************************************************************}
{*                                                                          *}
{*                              Program/Module                              *}
{*                                   from                                   *}
{*                 NUMERICAL TOOLBOX FOR VERIFIED COMPUTING I               *}
{*                         Basic Numerical Problems                         *}
{*                                                                          *}
{*      Copyright (c) 1993   Rolf Hammer, Matthias Hocks, Dietmar Ratz      *}
{*                                                                          *}
{* This program/module is free software for non-commercial use. For details *}
{* on theory, algorithms, and programs, see the book                        *}
{*                                                                          *}
{*  R. Hammer, M. Hocks, U. Kulisch, D. Ratz:  Numerical Toolbox for        *}
{*  Verified Computing I - Basic Numerical Problems. Springer-Verlag,       *}
{*  Heidelberg (ISBN 3-540-57118-3), New York (ISBN 0-387-57118-3), 1993.   *}
{*                                                                          *}
{* This program/module is distributed WITHOUT ANY WARRANTY. For details,    *}
{* see the "Disclaimer / Legal Matters" of the book (page iv).              *}
{*                                                                          *}
{****************************************************************************}
{----------------------------------------------------------------------------}
{ Purpose: Determination and enclosure of a root of a complex polynomial,    }
{    and of the coefficients of the deflated polynomial.                     }
{ Method: The root of a complex polynomial matches with the eigenvalue of    }
{    its companion matrix and the coefficients of the deflated polynomial    }
{    match to the components of the corresponding eigenvector. The eigen-    }
{    vector and eigenvalue are determined using the simplified Newton        }
{    iteration with iterative refinement.                                    }
{ Global procedures and functions:                                           }
{   procedure CPolyZero(...)      : computes an enclosure for a root and for }
{                                   the deflated complex polynomial          }
{   function CPolyZeroErrMsg(...) : delivers an error message text           }
{----------------------------------------------------------------------------}
module cpzero;     { Complex polynomial zero }
                   { -       -          ---- }
use
  r_util,          { Utilities of type real       }
  c_ari,           { Complex arithmetic           }
  ci_ari,          { Complex interval arithmetic  }
  cpoly,           { Complex polynomials          }
  cipoly;          { Complex interval polynomials }

{----------------------------------------------------------------------------}
{ Error codes used in this module. In the comments below p[0],..., p[n] are  }
{ the coefficients of a polynomial p.                                        }
{----------------------------------------------------------------------------}
const
  NoError   = 0;      { No error occurred.                             }
  ZeroPoly  = 1;      { Zero polynomial, i.e. n = 0 and p[0] = 0.      }
  ConstPoly = 2;      { Constant polynomial, i.e. n = 0 and p[0] <> 0. }
  InvFailed = 3;      { Inversion of the Jacobian failed.              }
  VerFailed = 4;      { Verified inversion of the Jacobian failed.     }
  IncFailed = 5;      { Inclusion failed.                              }
  IllProb   = 6;      { Illegal Problem, i.e. polynomials with lower   }
                      { bound <> 0 passed to the main procedure.       }

{----------------------------------------------------------------------------}
{ Error messages depending on the error code.                                }
{----------------------------------------------------------------------------}
global function CPolyZeroErrMsg(Err : integer) : string;
var
  Msg : string;
begin
  case Err of
    NoError  : Msg := '';
    ZeroPoly : Msg := 'Zero polynomial occurred';
    ConstPoly: Msg := 'Constant polynomial <> 0 occurred';
    InvFailed: Msg := 'Inversion of the Jacobian failed';
    VerFailed: Msg := 'Verified inversion of the Jacobian failed';
    IncFailed: Msg := 'Inclusion failed';
    IllProb  : Msg := 'Illegal polynomial with least index <> 0 occurred';
    else     : Msg := 'Code not defined';
  end;
  if (Err <> NoError) then Msg := 'Error: ' + Msg + '!';
  CPolyZeroErrMsg := Msg;
end;

function MaxNorm(p : CPolynomial) : real; { Function used to get the maximum }
var                                       { norm of 'p', where 'p' is        }
  i        : integer;                     { interpreted as a complex vector  }
  max, tmp : real;                        {----------------------------------}
begin
  max := abs(p[0]);
  for i := 1 to ub(p) do
  begin
    tmp := abs(p[i]);
    if (tmp > max) then max := tmp;
  end;
  MaxNorm := max;
end;

{----------------------------------------------------------------------------}
{ Purpose: Determination of a floating-point approximation for a root of     }
{    polynomial p and for the deflated polynomial.                           }
{ Parameters:                                                                }
{    In    : 'p'  : represents a complex polynomial by its coefficients      }
{    In/Out: 'z'  : starting approximation (in) for a root and is returned   }
{                   as an improved floating-point approximation              }
{    Out   : 'q'  : returns an approximation (for the coefficients of) the   }
{                   deflated polynomial                                      }
{            'Err': error flag                                               }
{ Description: For a given starting approximation 'z', a floating-point      }
{    Newton iteration is performed to get improved approximations for a root }
{    'z' of polynomial 'p', as well as for the coefficients 'q' of the       }
{    deflated polynomial.                                                    }
{----------------------------------------------------------------------------}
procedure Approximation(    p : CPolynomial; var z   : complex;
                        var q : CPolynomial; var Err : integer );
const
  kmax  = 50;                            { Maximum number of iteration steps }
  eps   = 10E-10;                        { Relative error of approximation   }
var
  k, i, n     : integer;
  t           : complex;
  delta, d, w : CPolynomial[0..ub(p)-1];
  qHlp        : CPolynomial[0..ub(p)-2];
begin
  Err := NoError;  n := ub(p);                              { Initialization }
                                                            {----------------}
  q[n-1] := p[n];                          { Determination of an approximate }
  for i := n-2 downto 0 do                 { eigenvector q                   }
    q[i] := q[i+1]*z + p[i+1];             {---------------------------------}

  k := 0;                                         { Floating-point iteration }
  repeat                                          { of z and q               }
                                                  {--------------------------}

    d[0] := #*( -z*q[0] - p[0] );                { Computation of the defect }
    for i := 1 to n-1 do                         { d = (A - z*I)*q           }
      d[i] := #*( q[i-1] - z*q[i] - p[i] );      {---------------------------}

    w[n-1] := q[n-1];                        { Computation of R, the inverse }
    for i := n-2 downto 0 do                 { of the Jacobian J             }
      w[i] := #*( q[i] + z*w[i+1] );         {-------------------------------}

    if (w[0] = 0) then
      Err := InvFailed
    else
      begin
        delta[n-1] := d[n-1];                 { Computation of delta^(k+1) = }
        for i := n-2 downto 0 do              { g(delta^(k))                 }
          delta[i] := d[i] + z*delta[i+1];    {------------------------------}
        t := delta[0]/w[0];
        for i := 0 to n-2 do 
          delta[i] := -delta[i+1] + t*w[i+1];
        delta[n-1] := t;

        for i := 0 to n-2 do         { Determine the new iterates of q and z }
        begin                        { and store the first n-1 components of }
          q[i] := q[i] + delta[i];   { q in qHlp                             }
          qHlp[i] := q[i];           {---------------------------------------}
        end;
        z := z + delta[n-1];

        k := k+1;
      end;

  { Stop iteration if the relative round-off error of the coefficients and }
  { of the zero is approximately <= a given epsilon. Stop also if the      }
  { maximum number of iteration steps is exceeded, or if an error occurred.}
  until ( Err <> NoError) 
        or ( k = kmax ) 
        or ( (MaxNorm(delta) / max(MaxNorm(qHlp),abs(z))) <= eps);

end; { procedure Approximation }

{----------------------------------------------------------------------------}
{ Purpose: Determination of enclosures for a root of polynomial p and for    }
{    the deflated polynomial.                                                }
{ Parameters:                                                                }
{    In    : 'p'  : represents a complex polynomial by its coefficients      }
{            'q'  : approximation (of the coefficients of) the deflated      }
{                   polynomial                                               }
{            'z'  : floating-point approximation of a root                   }
{    Out   : 'qq' : returns an enclosure (for the coefficients of) the       }
{                   deflated polynomial                                      }
{            'zz' : returns an enclosure of the root                         }
{            'Err': error flag                                               }
{ Description: For starting floating-point approximations 'z' of a root      }
{    and 'q' of the deflated polynomial an interval residual iteration is    }
{    performed to get enclosures 'zz' for the root of polynomial 'p', as     }
{    well as for the coefficients of the deflated polynomial.                }
{----------------------------------------------------------------------------}
procedure IntervalIteration( p, q   : CPolynomial;      z   : complex;
                            var qq  : CIPolynomial; var zz  : cinterval;
                            var Err : integer                           );
const
  kmax   = 10;                           { Maximum number of iteration steps }
var
  i, k, n         : integer;
  eps             : real;
  tt              : cinterval;
  dd, vv, ww      : CIPolynomial[0..ub(p)-1];
  ddelta, ddelta_0, 
  ddelta_old      : CIPolynomial[0..ub(p)-1];
begin
  Err := NoError;  n := ub(p);                              { Initialization }
                                                            {----------------}
  dd[0] := ##( -p[0] - z*q[0] );               { Compute an inclusion of the }
  for i := 1 to n-1 do                         { defect [d] = (A - z*I)*q    }
    dd[i] := ##( q[i-1] - z*q[i] - p[i]);      {-----------------------------}

  ww[n-1] := q[n-1];                      { Compute an inclusion of [R], the }
  for i := n-2 downto 0 do                { inverse of the Jacobian J        }
    ww[i] := ##( q[i] + z*ww[i+1] );      {----------------------------------}
  if (0 in ww[0]) then
    Err := VerFailed
  else
    begin
      ddelta_0[n-1] := dd[n-1];                         { Compute a starting }
      for i := n-2 downto 0 do                          { inclusion [y]^(0)  }
        ddelta_0[i] := ##( dd[i] + z*ddelta_0[i+1] );   {--------------------}
      tt := ddelta_0[0]/ww[0];
      for i := 0 to n-2 do 
        ddelta_0[i] := ##( -ddelta_0[i+1] + tt*ww[i+1] );
      ddelta_0[n-1] := tt;

      { Interval iteration }
      k := 0;  ddelta := ddelta_0;                      { Interval iteration }
      repeat                                            {--------------------}
        ddelta_old := ddelta;

        case k of                                     { Slightly enlarge the }
          0..3:  eps := 0.125;                        { inclusion interval   }
          4..6:  eps := 0.5;                          {----------------------}
          else:  eps := 5;
        end;
        ddelta_old := blow(ddelta,eps);

        vv[n-1] := 0;                   { Determine a new inclusion interval }
        for i := n-2 downto 0 do        {------------------------------------}
          vv[i] := ##( ddelta_old[n-1]*ddelta_old[i] + z*vv[i+1] );
        vv[0] := vv[0]/ww[0];
        for i := 0 to n-2 do 
          ddelta[i] := ##( ddelta_0[i] + vv[i+1] - vv[0]*ww[i+1] );
        ddelta[n-1] := ddelta_0[n-1] - vv[0];

        k := k+1;
      until (ddelta in ddelta_old) or (k = kmax);

      if (ddelta in ddelta_old) then            { Verification of the result }
        begin                                   {----------------------------}
          for i := 0 to n-2 do qq[i] := ##( q[i] + ddelta[i] );
          qq[n-1] := p[n];
          zz := ##( z + ddelta[n-1] );
        end
      else
        Err := IncFailed;
    end;

end;  { procedure IntervalIteration }

{----------------------------------------------------------------------------}
{ Purpose: Determination of enclosures for a root of polynomial p and for    }
{    the deflated polynomial                                                 }
{ Parameters:                                                                }
{    In    : 'p'  : represents a complex polynomial by its coefficients      }
{            'z'  : floating-point approximation of a root                   }
{    Out   : 'qq' : returns an enclosure (for the coefficients of) the       }
{                   deflated polynomial                                      }
{            'zz' : returns an enclosure of the root                         }
{            'Err': error flag                                               }
{ Description: For starting approximations 'z' of a root, enclosures 'zz'   }
{    for a root of polynomial 'p', as well as 'qq' for the coefficients of   }
{    the deflated polynomial are computed. Since the root of a complex       }
{    polynomial matches with an eigenvalue of its companion matrix and the   }
{    coefficients of the deflated polynomial match to the components of the  }
{    corresponding eigenvector, the eigenvector and eigenvalue are determined}
{    by iterative refinement. First, good floating-point approximations are  }
{    computed, and then verified enclosures for a root and for a deflated    }
{    polynomial are returned.                                                }
{----------------------------------------------------------------------------}
global procedure CPolyZero(    p   : CPolynomial;
                               z   : complex;
                           var qq  : CIPolynomial;
                           var zz  : cinterval;
                           var Err : integer      );
var
  i, n : integer;
  q    : CPolynomial[0..ub(p)-1];
begin
  if (lb(p) <> 0) or (lb(qq) <> 0) then     { Illegal polynomial declaration }
    Err := IllProb
  else
    begin
      Err := NoError;  n := ub(p);                          { Initialization }

      if (n = 0) or (n = 1) then
        if (n = 0) then                         { Polynomial of degree n = 0 }
          if (p[0] = 0) then Err := ZeroPoly    { Zero polynomial            }
                        else Err := ConstPoly   { Constant polynomial        }
        else                                    { Polynomial of degree n = 1 }
          if (p[1] = 0) then
            if (p[0] = 0) then Err := ZeroPoly  { Zero polynomial            }
                          else Err := ConstPoly { Constant polynomial        }
          else
            begin
              zz := -intval(p[0]) / p[1];
              qq[0] := p[1];
            end
      else                                      { Polynomial of degree n > 1 }
        begin                                   { (Common case)              }
          Approximation(p,z,q,Err);
          if (Err = NoError) then IntervalIteration(p,q,z,qq,zz,Err);
        end;
    end;
end; { procedure CPolyZero }

{----------------------------------------------------------------------------}
{ Module initialization part                                                 }
{----------------------------------------------------------------------------}
begin
  { Nothing to initialize }
end.
