{****************************************************************************}
{*                                                                          *}
{*                              Program/Module                              *}
{*                                   from                                   *}
{*                 NUMERICAL TOOLBOX FOR VERIFIED COMPUTING I               *}
{*                         Basic Numerical Problems                         *}
{*                                                                          *}
{*      Copyright (c) 1993   Rolf Hammer, Matthias Hocks, Dietmar Ratz      *}
{*                                                                          *}
{* This program/module is free software for non-commercial use. For details *}
{* on theory, algorithms, and programs, see the book                        *}
{*                                                                          *}
{*  R. Hammer, M. Hocks, U. Kulisch, D. Ratz:  Numerical Toolbox for        *}
{*  Verified Computing I - Basic Numerical Problems. Springer-Verlag,       *}
{*  Heidelberg (ISBN 3-540-57118-3), New York (ISBN 0-387-57118-3), 1993.   *}
{*                                                                          *}
{* This program/module is distributed WITHOUT ANY WARRANTY. For details,    *}
{* see the "Disclaimer / Legal Matters" of the book (page iv).              *}
{*                                                                          *}
{****************************************************************************}
{----------------------------------------------------------------------------}
{ Purpose: Computation of an approximate inverse of a real square matrix.    }
{ Method:  LU decomposition applying Crout's algorithm.                      }
{ Global functions and procedures:                                           }
{   procedure MatInv(...)      : Matrix inversion                            }
{   function MatInvErrMsg(...) : To get an error message text                }
{----------------------------------------------------------------------------}
module matinv;    { Matrix inversion routines }
                  { ---    ---                }
use
  mv_ari;         { Real matrix/vector arithmetic }

{----------------------------------------------------------------------------}
{ Error codes used in this module.                                           }
{----------------------------------------------------------------------------}
const
  NoError      = 0;     { No error occurred.                             }
  NotSquare    = 1;     { Matrix to be inverted is not square.           }
  DimensionErr = 2;     { Input and output matrices differ in shape.     }
  Singular     = 3;     { Matrix to be inverted is probably singular.    }

{----------------------------------------------------------------------------}
{ Error messages depending on the error code.                                }
{----------------------------------------------------------------------------}
global function MatInvErrMsg ( Err : integer ) : string;
var
  Msg : string;
begin
  case Err of
    NoError     : Msg := '';
    NotSquare   : Msg := 'Matrix to be inverted is not square';
    DimensionErr: Msg := 'Input and output matrices differ in shape';
    Singular    : Msg := 'Inversion failed, matrix is probably singular';
    else        : Msg := 'Code not defined';
  end;
  if (Err <> NoError) then Msg := 'Error: ' + Msg + '!';
  MatInvErrMsg := Msg;
end;

{----------------------------------------------------------------------------}
{ Purpose: The procedure 'MatInv' may be used for the computation of an      }
{    approximate inverse of a real square matrix.                            }
{ Parameters:                                                                }
{    In  : 'A'   : matrix to be inverted, passed as reference parameter to   }
{                  save computation time for copying 'A'.                    }
{    Out : 'R'   : approximate inverse.                                      }
{          'Err' : error code.                                               }
{ Description:                                                               }
{   Inversion of a regular matrix A stored in 'A' using LU decomposition.    }
{   For LU decomposition, formally a permutation matrix P is determined so   }
{   that the product P*A may be decomposed into a lower-triangular matrix L  }
{   and an upper-triangular matrix U with P*A = L*U. The diagonal elements   }
{   of L are set to 1. Using Crout's algorithm, the elements of both matri-  }
{   ces L and U are stored by temporary overwriting a copy of the input      }
{   matrix 'A'. The permutation matrix P is not explicitly generated. The    }
{   indices of row interchanges are stored in the index vector 'p' instead.  }
{   The i-th element of P*b may be accessed indirectly using the p[i]-th     }
{   entry of 'b'. Accurate expressions are used to avoid cancellation errors.}
{   The k-th column of the inverse R of P*A is computed by forward/backward  }
{   substitution with the k-th unit vector e_k as the right-hand side of the }
{   system: U*y = P*e_k ==> y, L*x = y ==> x. For error codes, see above.    }
{----------------------------------------------------------------------------}
global procedure MatInv ( var A   : rmatrix;
                          var R   : rmatrix;
                          var Err : integer  );
const
  Tiny = 1E-200;     { A divisor less than 'Tiny' is handled as zero }
var
  n1, nn, m1, mm   : integer;    { For lower and upper bounds of 'A' }
  p1, pn, q1, qm   : integer;    { For lower and upper bounds of 'R' }
  i, j, k, l, n, m : integer;    { For loops      }
  Max, Temp        : real;       { Help variables }
  AA        : rmatrix[1..ub(A)-lb(A)+1,1..ub(A)-lb(A)+1];   { A copy of 'A'}
  p         : dynamic array[1..ub(A)-lb(A)+1] of integer;
  v, x      : rvector[1..ub(A)-lb(A)+1];
begin
  { Get lower and upper bounds of the rows and columns of 'A' and 'R' }
  n1 := lb(A,1);  nn := ub(A,1);  n := nn-n1+1;
  m1 := lb(A,2);  mm := ub(A,2);  m := mm-m1+1;
  p1 := lb(R,1);  pn := ub(R,1);
  q1 := lb(R,2);  qm := ub(R,2);

  { Check for correct dimensions of 'A' and 'R' }
  Err := NoError;

  if (n <> m) then                        { Error: 'A' is not square         }
    Err := NotSquare
  else if ( (n <> pn-p1+1) or
            (m <> qm-q1+1) ) then         { Error: Dimensions not compatible }
    Err := DimensionErr
  else if (n = 2) then                    { Special case: (2,2)-matrix       }
    begin
      AA := A;      { Copy 'A' to avoid computation of index offsets }
      Temp := #*( AA[1,1]*AA[2,2] - AA[2,1]*AA[1,2] );   { = Det(AA) }
      if (abs(Temp) < Tiny) then
        Err := Singular
      else
        begin
          R[p1,q1] :=  AA[2,2] / Temp;  R[p1,qm] := -AA[1,2] / Temp;
          R[pn,q1] := -AA[2,1] / Temp;  R[pn,qm] :=  AA[1,1] / Temp;
        end;
    end
  else   { Usual case: Dimension of 'A' > 2 }
    begin
      AA := A;         { Copy 'A' to avoid computation of index offsets }
      for i := 1 to n do p[i] := i;   { Initializing permutation vector }

      { Start LU factorization }
      i := 1;
      while ((Err = NoError) and (i <= n)) do
      begin
        { Compute the numerators of those elements of the i-th column }
        { of L which are not updated so far and store them in 'v'.    }
        for k := i to n do
          v[k] := #*(AA[k,i] - for j:=1 to i-1 sum(AA[k,j] * AA[j,i]));

        { Look for the column pivot }
        j := i;  Max := abs(v[i]);
        for k := i+1 to n do
        begin
          Temp := abs(v[k]);
          if (Temp > Max) then
            begin  j := k;  Max := Temp;  end;
        end;

        { Swap rows of 'AA' and 'v', store the permutation in 'p' }
        if (j <> i) then
        begin
          x := AA[i];    AA[i] := AA[j];  AA[j] := x;
          k := p[i];     p[i] := p[j];    p[j] := k;
          Temp := v[i];  v[i] := v[j];    v[j] := Temp;
        end;

        if (Max < Tiny) then     { Pivot element < 'Tiny', inversion failed }
          Err := Singular        { matrix 'A' assumed to be singular        }
        else
          begin
            Temp := v[i];
            AA[i,i] := Temp;      { Update the diagonal element of U }
          end;

        if (Err = NoError) then
        begin
          { Update U's row and L's column elements }
          for k := i+1 to n do
          begin
            AA[i,k] := #*(AA[i,k] - for j:=1 to i-1 sum(AA[i,j]*AA[j,k]));
            AA[k,i] := v[k] / Temp;
          end;
          i := i+1;
        end;
      end; {while}

      if (Err = NoError) then

        { Now 'AA' is overwritten with the subdiagonal elements of L in its  }
        { lower left triangle and with the elements of U in its diagonal and }
        { its upper right triangle. The elements of the inverse matrix are   }
        { computed column by column using forward/backward substitution.     }
        {--------------------------------------------------------------------}
        for k := 1 to n do
        begin
          { Forward substitution: L*x = P*e_k, where e_k is the k-th unit }
          { vector. Note: If P*e_k has m leading zeros, this results in   }
          { x_i = 0 for 1,..,l-1 and x_l = 1. Thus, forward substitution  }
          { starts at index l+1.                                          }
          {---------------------------------------------------------------}
          l := 1;
          while (p[l] <> k) do
            begin  x[l] := 0; l := l+1;  end;
          x[l] := 1;
          for i := l+1 to n do
            x[i] := -#*(for j:=l to i-1 sum(AA[i,j]*x[j]));

          { Backward substitution: U * x = x, where the right-hand side is }
          { the result of the forward substitution. It will be overwritten }
          { by the solution of the system, the k-th column of the inverse  }
          { matrix.                                                        }
          {----------------------------------------------------------------}
          for i := n downto 1 do
            x[i] := #*(x[i] - for j:=i+1 to n sum(AA[i,j]*x[j])) / AA[i,i];

          R[*,q1+k-1] := x;   { Remember index offset! }
      end; {if (Err = NoError) ...}
    end; {Usual case}
end;

{----------------------------------------------------------------------------}
{ Module initialization part                                                 }
{----------------------------------------------------------------------------}
begin
  { Nothing to initialize }
end.
