{****************************************************************************}
{*                                                                          *}
{*                              Program/Module                              *}
{*                                   from                                   *}
{*                 NUMERICAL TOOLBOX FOR VERIFIED COMPUTING I               *}
{*                         Basic Numerical Problems                         *}
{*                                                                          *}
{*     Copyright (c) 1993-96 Rolf Hammer, Matthias Hocks, Dietmar Ratz      *}
{*                                                                          *}
{* This program/module is free software for non-commercial use. For details *}
{* on theory, algorithms, and programs, see the book                        *}
{*                                                                          *}
{*  R. Hammer, M. Hocks, U. Kulisch, D. Ratz:  Numerical Toolbox for        *}
{*  Verified Computing I - Basic Numerical Problems. Springer-Verlag,       *}
{*  Heidelberg (ISBN 3-540-57118-3), New York (ISBN 0-387-57118-3), 1993.   *}
{*                                                                          *}
{* This program/module is distributed WITHOUT ANY WARRANTY. For details,    *}
{* see the "Disclaimer / Legal Matters" of the book (page iv).              *}
{*                                                                          *}
{****************************************************************************}
{----------------------------------------------------------------------------}
{ Purpose: Definition of an extended interval arithmetic which allows the    }
{    division by an interval containing zero.                                }
{ Method: Overloading of operators for arithmetic and lattice operations     }
{    of data type 'xinterval'.                                               }
{ Global types, operators, and functions:                                    }
{    types      KindType         : component type of extended intervals      }
{               xinterval        : data type for extended intervals          }
{    operators  div, -, **       : operators of extended interval arithmetic }
{    function   EmptyIntval      : delivers empty set as irregular interval  }
{ Updates:                                                                   }
{    04.03.1996 : Modification of extended interval operations and kind type }
{----------------------------------------------------------------------------}
module xi_ari;

use i_ari;  { Interval arithmetic }

{----------------------------------------------------------------------------}
{ Global type definitions                                                    }
{----------------------------------------------------------------------------}
{ An extended interval 'x', represented by the type 'xinterval', is defined  }
{ according to the following rules (a <= b):                                 }
{                                                                            }
{ x = [a, b]              :  x.kind = Finite,     x.inf = a, x.sup = b       }
{ x = [a, +oo]            :  x.kind = PlusInfty,  x.inf = a, x.sup undefined }
{ x = [-oo, a]            :  x.kind = MinusInfty, x.sup = a, x.inf undefined }
{ x = [-oo, a] v [b, +oo] :  x.kind = Double,     x.inf = b, x.sup = a       }
{ x = [-oo, +oo]          :  x.kind = Double,     x.inf = a, x.sup = a       }
{ x = [/]                 :  x.kind = Empty,      x.inf  and x.sup undefined }
{                                                                            }
{ In this definition, 'v' stands for the set union and 'oo' for infinity.    }
{----------------------------------------------------------------------------}
global type
  KindType     = (Finite, PlusInfty, MinusInfty, Double, Empty);
  xinterval    = global record                         { Extended intervals  }
                          kind     : KindType;         { according to the    }
                          inf, sup : real;             { definition above    }
                        end;                           {---------------------}

{----------------------------------------------------------------------------}
{ Function 'EmptyIntval' delivers an empty interval (empty set)              }
{----------------------------------------------------------------------------}
global function EmptyIntval : interval;
begin
  EmptyIntval.inf :=  999999999;      { Definition of an irregular interval  }
  EmptyIntval.sup := -999999999;      { EmptyIntval = [999999999,-999999999] }
end;                                  {--------------------------------------}

{----------------------------------------------------------------}
{ Extended interval division 'A / B', where 0 in 'B' is allowed. }
{----------------------------------------------------------------}
global operator div (A, B : interval) quotient : xinterval;
  var
    C : interval;
    Q : xinterval;
  begin
    if (0 in B) then { extended interval division }
      begin
        if (0 in A) then
          begin                       { Q = [-oo, +oo] = [-oo, 0] v [0, +oo] }
            Q.kind:= Double;          {--------------------------------------}
            Q.sup := 0;
            Q.inf := 0;
          end
        else if (0 = B) then
          begin                                                    { Q = [/] }
            Q.kind := Empty;                                       {---------}
          end
        else if ((sup(A) < 0) and (sup(B) = 0)) then
          begin                                           { Q = [Q.inf, +oo] }
            Q.kind:= PlusInfty;                           {------------------}
            Q.inf := sup(A) /< inf(B);
          end
        else if ((sup(A) < 0) and (inf(B) < 0) and (sup(B) > 0)) then
          begin                            { Q = [-oo, Q.sup] v [Q.inf, +oo] }
            Q.kind:= Double;               {---------------------------------}
            Q.sup := sup(A) /> sup(B);
            Q.inf := sup(A) /< inf(B);
          end
        else if ((sup(A) < 0) and (inf(B) = 0)) then
          begin                                           { Q = [-oo, Q.sup] }
            Q.kind:= MinusInfty;                          {------------------}
            Q.sup := sup(a) /> sup(b);
          end
        else if ((inf(A) > 0) and (sup(B) = 0)) then
          begin                                           { Q = [-oo, Q.sup] }
            Q.kind:= MinusInfty;                          {------------------}
            Q.sup := inf(A) /> inf(B);
          end
        else if ((inf(A) > 0) and (inf(B) < 0) and (sup(B) > 0)) then
          begin                            { Q = [-oo, Q.sup] v [Q.inf, +oo] }
            Q.kind:= Double;               {---------------------------------}
            Q.sup := inf(A) /> inf(B);
            Q.inf := inf(A) /< sup(B);
          end
        else if ((inf(A) > 0) and (inf(B) = 0)) then
          begin                                           { Q = [Q.inf, +oo] }
            Q.kind:= PlusInfty;                           {------------------}
            Q.inf := inf(A) /< sup(B);
          end
      end { 0 in B }
    else  { not (0 in B) ==> standard interval division }
      begin                                             { Q = [C.inf, C.sup] }
        C     := A / B;                                 {--------------------}
        Q.kind:= Finite;
        Q.inf := C.inf;
        Q.sup := C.sup;
      end;
    quotient := Q;
  end;

{----------------------------------------------------------------}
{ Subtraction of an extended interval 'B' from a real value 'a'. }
{----------------------------------------------------------------}
global operator - (a : real; B : xinterval) difference : xinterval;
  var
    D : xinterval;
  begin
    case B.kind of
      Finite     : begin                                { D = [D.inf, D.sup] }
                      D.kind:= Finite;                  {--------------------}
                      D.inf := a -< B.sup;
                      D.sup := a -> B.inf;
                   end;
      PlusInfty  : begin                                    { D = [inf, +oo] }
                      D.kind:= MinusInfty;                  {----------------}
                      D.sup := a -> B.inf;
                   end;
      MinusInfty : begin                                    { D = [-oo, sup] }
                      D.kind:= PlusInfty;                   {----------------}
                      D.inf := a -< B.sup;
                   end;
      Double     : begin                   { D = [-oo, D.sup] v [D.inf, +oo] }
                      D.kind:= Double;     {---------------------------------}
                      D.inf := a -< B.sup;
                      D.sup := a -> B.inf;
                      if (D.inf < D.sup) then
                        D.inf := D.sup;
                   end;
      Empty      : begin                                           { D = [/] }
                      D.kind:= Empty;                              {---------}
                   end;
    end;
    difference := D;
  end;

{----------------------------------------------------------------------------}
{ Intersection of an interval 'X' and an extended interval 'Y'. The result   }
{ is given as a pair (vector) of intervals, where one or both of them can    }
{ be empty intervals ('EmptyIntval').                                        }
{----------------------------------------------------------------------------}
global operator ** (X: interval; Y: xinterval) Intersect : ivector[1..2];
  var
    H : interval;
  begin
    Intersect[1]:= EmptyIntval;
    Intersect[2]:= EmptyIntval;

    case Y.kind of
      Finite    : begin                       { [X.inf,X.sup]**[Y.inf,Y.sup] }
                    H := intval(Y.inf,Y.sup); {------------------------------}
                    if not (X >< H) then
                      Intersect[1]:= X ** H;
                  end;
      PlusInfty : if (X.sup >= Y.inf) then    { [X.inf,X.sup] ** [Y.inf,+oo] }
                  begin                       {------------------------------}
                    if (X.inf > Y.inf) then
                      Intersect[1]:= X
                    else
                      Intersect[1]:= intval(Y.inf, X.sup);
                  end;
      MinusInfty: if (Y.sup >= X.inf) then    { [X.inf,X.sup] ** [-oo,Y.sup] }
                  begin                       {------------------------------}
                    if (X.sup < Y.sup) then
                      Intersect[1]:= X
                    else
                      Intersect[1]:= intval(X.inf, Y.sup)
                  end;
      Double    : if ((X.inf <= Y.sup) and (Y.inf <= X.sup)) then
                    begin
                      Intersect[1]:= intval(X.inf, Y.sup);  { X**[-oo,Y.sup] }
                      Intersect[2]:= intval(Y.inf, X.sup);  { X**[Y.inf,+oo] }
                    end                                    {----------------}
                  else if (Y.inf <= X.sup) then
                    begin                       { [X.inf,X.sup]**[Y.inf,+oo] }
                      if (X.inf >= Y.inf) then  {----------------------------}
                        Intersect[1]:= X
                      else
                        Intersect[1]:= intval(Y.inf, X.sup);
                    end
                  else if (X.inf <= Y.sup) then
                    begin                       { [X.inf,X.sup]**[-oo,Y.sup] }
                      if (X.sup <= Y.sup) then  {----------------------------}
                        Intersect[1]:=  X
                      else
                        Intersect[1]:=  intval(X.inf, Y.sup);
                    end;
      Empty     : ;                                   { [X.inf,X.sup] ** [/] }
    end; { CASE kind OF ... }                         {----------------------}
  end; { OPERATOR ** ... }

{----------------------------------------------------------------------------}
{ Module initialization part                                                 }
{----------------------------------------------------------------------------}
begin
  { Nothing to initialize }
end.
